using System;
using System.Reflection;
using Cairo;
using HarmonyLib;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.MathTools;
using Vintagestory.GameContent;

namespace AltMapIconRenderer {
	public class AltMapIconRendererSystem: ModSystem {
		public const string harmonyId = "altmapiconrenderer";

		public const float waypointScale = 0.5f;
		public const float playerScale = 0.75f;
		public const float hoverSize = 8;

		public const float zoomMin = 0.75f;
		public const float zoomMax = 1f;

		public const float outlineOffsetDefault = 0.2f;
		public const float outlineOffsetThin = 0.1f;
		public const float squareOutlineScaleDefault = 1.2f;
		public const float squareOutlineScaleThin = 1.1f;

		public const string boolWordTrue = "on";
		public const string boolWordFalse = "off";
		public static readonly string[] boolWords = new string[] { boolWordTrue, boolWordFalse };

		public const string thinWordTrue = "thin";
		public const string thinWordFalse = "thick";
		public static readonly string[] thinWords = new string[] { thinWordTrue, thinWordFalse };

		public static Vec4f rgbaOutline = new Vec4f(0, 0, 0, 1f);
		public static Vec4f rgbaHover = new Vec4f(0.75f, 0, 0, 1f);
		public static float outlineOffset;
		public static float squareOutlineScale;
		public static bool hideWaypoints;

		public static AltMapIconRendererConfigManager config;
		public static LoadedTexture squareTex;
		public static Vec4f playerColour;

		protected ICoreClientAPI capi;
		protected Harmony harmony;

		public override void StartClientSide(ICoreClientAPI capi) {
			base.StartClientSide(capi);
			this.capi = capi;

			AltMapIconRendererSystem.config = new AltMapIconRendererConfigManager(this.capi);
			AltMapIconRendererSystem.playerColour = new Vec4f();
			AltMapIconRendererSystem.hideWaypoints = false;

			this.generateSquareTexture();

			this.SetPlayerColour(AltMapIconRendererSystem.config.Get().player_colour);
			this.SetThinOutlines(AltMapIconRendererSystem.config.Get().thin_outlines);

			this.capi.ChatCommands.Create("amir")
					.BeginSubCommand("square")
							.WithDescription("Render waypoint icons as squares.")
							.WithArgs(this.capi.ChatCommands.Parsers.WordRange("off/on", AltMapIconRendererSystem.boolWords))
							.HandleWith(this.cmdSquare)
					.EndSubCommand()
					.BeginSubCommand("outline")
							.WithDescription("Set waypoint icon outline style.")
							.WithArgs(this.capi.ChatCommands.Parsers.WordRange("thin/thick", AltMapIconRendererSystem.thinWords))
							.HandleWith(this.cmdOutline)
					.EndSubCommand()
					.BeginSubCommand("pc")
							.WithDescription("Set the player pin's colour.")
							.WithArgs(this.capi.ChatCommands.Parsers.Word("hex"))
							.HandleWith(this.cmdPc)
					.EndSubCommand()
					.BeginSubCommand("hide")
							.WithDescription("Hide waypoint icons.")
							.HandleWith(this.cmdHide)
					.EndSubCommand()
					.BeginSubCommand("show")
							.WithDescription("Show waypoint icons.")
							.HandleWith(this.cmdShow)
					.EndSubCommand()
					.BeginSubCommand("help")
							.WithDescription("Show Alternative Map Icon Renderer command help.")
							.HandleWith(this.cmdHelp)
					.EndSubCommand();

			this.harmony = new Harmony(AltMapIconRendererSystem.harmonyId);

			// Waypoint icons.
			this.harmony.Patch(typeof(WaypointMapComponent).GetMethod(nameof(WaypointMapComponent.Render), BindingFlags.Public | BindingFlags.Instance),
					prefix: new HarmonyMethod(typeof(AltWaypointMapComponent).GetMethod(nameof(AltWaypointMapComponent.Render))));
			this.harmony.Patch(typeof(WaypointMapComponent).GetMethod(nameof(WaypointMapComponent.OnMouseMove), BindingFlags.Public | BindingFlags.Instance),
					prefix: new HarmonyMethod(typeof(AltWaypointMapComponent).GetMethod(nameof(AltWaypointMapComponent.OnMouseMove))));
			this.harmony.Patch(typeof(WaypointMapComponent).GetMethod(nameof(WaypointMapComponent.OnMouseUpOnElement), BindingFlags.Public | BindingFlags.Instance),
					prefix: new HarmonyMethod(typeof(AltWaypointMapComponent).GetMethod(nameof(AltWaypointMapComponent.OnMouseUpOnElement))));

			// Player icon.
			this.harmony.Patch(typeof(EntityMapComponent).GetMethod(nameof(EntityMapComponent.Render), BindingFlags.Public | BindingFlags.Instance),
					prefix: new HarmonyMethod(typeof(AltEntityMapComponent).GetMethod(nameof(AltEntityMapComponent.Render))));
		}

		public override void Dispose() {
			base.Dispose();

			if(AltMapIconRendererSystem.squareTex != null) {
				AltMapIconRendererSystem.squareTex.Dispose();
			}

			this.harmony.UnpatchAll(AltMapIconRendererSystem.harmonyId);
		}

		public void SetPlayerColour(string hex) {
			try {
				double[] colour = ColorUtil.Hex2Doubles(hex);
				AltMapIconRendererSystem.playerColour.R = (float)colour[0];
				AltMapIconRendererSystem.playerColour.G = (float)colour[1];
				AltMapIconRendererSystem.playerColour.B = (float)colour[2];
				if(colour.Length >= 4) {
					AltMapIconRendererSystem.playerColour.A = (float)colour[3];
				}
			} catch(Exception) {
				// Default colour.
				AltMapIconRendererSystem.playerColour.R = 1;
				AltMapIconRendererSystem.playerColour.G = 1;
				AltMapIconRendererSystem.playerColour.B = 1;
				AltMapIconRendererSystem.playerColour.A = 1;
			}
		}

		public void SetThinOutlines(bool thin) {
			AltMapIconRendererSystem.outlineOffset = thin
					? AltMapIconRendererSystem.outlineOffsetThin
					: AltMapIconRendererSystem.outlineOffsetDefault;
			AltMapIconRendererSystem.squareOutlineScale = thin
					? AltMapIconRendererSystem.squareOutlineScaleThin
					: AltMapIconRendererSystem.squareOutlineScaleDefault;
		}

		protected TextCommandResult cmdSquare(TextCommandCallingArgs args) {
			string square = (string)args[0];

			bool b;
			switch(square) {
				case AltMapIconRendererSystem.boolWordTrue: {
					b = true;
					break;
				}
				default: {
					b = false;
					break;
				}
			}

			AltMapIconRendererSystem.config.Get().square_waypoints = b;
			AltMapIconRendererSystem.config.Save();
			
			return TextCommandResult.Success(print("Square waypoints " + (b ? "enabled" : "disabled") + "."));
		}

		protected TextCommandResult cmdOutline(TextCommandCallingArgs args) {
			string thin = (string)args[0];

			bool b;
			switch(thin) {
				case AltMapIconRendererSystem.thinWordTrue: {
					b = true;
					break;
				}
				default: {
					b = false;
					break;
				}
			}

			this.SetThinOutlines(b);

			AltMapIconRendererSystem.config.Get().thin_outlines = b;
			AltMapIconRendererSystem.config.Save();
			
			return TextCommandResult.Success(print("Thin outlines " + (b ? "enabled" : "disabled") + "."));
		}

		protected TextCommandResult cmdPc(TextCommandCallingArgs args) {
			string hex = (string)args[0];

			this.SetPlayerColour(hex);

			AltMapIconRendererSystem.config.Get().player_colour = vec4f2hex(AltMapIconRendererSystem.playerColour);
			AltMapIconRendererSystem.config.Save();

			return TextCommandResult.Success(print("Player colour set."));
		}

		protected TextCommandResult cmdHide(TextCommandCallingArgs args) {
			AltMapIconRendererSystem.hideWaypoints = true;

			return TextCommandResult.Success(print("Icons hidden."));
		}

		protected TextCommandResult cmdShow(TextCommandCallingArgs args) {
			AltMapIconRendererSystem.hideWaypoints = false;

			return TextCommandResult.Success(print("Icons shown."));
		}

		protected TextCommandResult cmdHelp(TextCommandCallingArgs args) {
			string message = "[Alternative Map Icon Renderer] Instructions:\n"
					+ "To toggle square waypoints, use:\t<code>.amir square [off, on]</code>\n"
					+ "To modify the waypoint outline style, use:\t<code>.amir outline [thick, thin]</code>\n"
					+ "To modify the player pin colour, use:\t<code>.amir pc [#colour]</code>\n"
					+ "To temporarily hide waypoint icons, use:\t<code>.amir hide</code>\n"
					+ "To show waypoint icons again, use:\t<code>.amir show</code>";

			return TextCommandResult.Success(message);
		}

		protected static string print(string text) {
			return "[Alternative Map Icon Renderer] " + text;
		}

		public static string vec4f2hex(Vec4f colour) {
			if(colour.A != 1) {
				// 8-digit hex code.
				return String.Format("#{0:X2}{1:X2}{2:X2}{3:X2}", (int)(255 * colour.R), (int)(255 * colour.G), (int)(255 * colour.B), (int)(255 * colour.A));
			} else {
				// 6-digit hex code.
				return String.Format("#{0:X2}{1:X2}{2:X2}", (int)(255 * colour.R), (int)(255 * colour.G), (int)(255 * colour.B));
			}
		}

		protected void generateSquareTexture() {
			AltMapIconRendererSystem.squareTex = new LoadedTexture(this.capi);
			ImageSurface surface = new ImageSurface(Format.Argb32, 20, 20);
			Context context = new Context(surface);

			context.SetSourceRGBA(1, 1, 1, 1);
			context.Rectangle(0, 0, 20, 20);
			context.Fill();
			
			this.capi.Gui.LoadOrUpdateCairoTexture(surface, true, ref AltMapIconRendererSystem.squareTex);
			context.Dispose();
			surface.Dispose();
		}
	}
}